/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2003 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Marcus Boerger <helly@php.net>                               |
  +----------------------------------------------------------------------+
*/

/* $Id: util.c,v 1.4 2004/04/27 17:12:06 boerger Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"

#include "zend_compile.h"
#include "zend_interfaces.h"
#include "zend_exceptions.h"

#include "fopen_wrappers.h"
#include "ext/standard/basic_functions.h"
#include "ext/standard/php_filestat.h"

#include "php_util.h"

/* declare the class entry */
extern zend_class_entry *zend_ce_iterator;


/* forward declaration for all methods use (class-name, method-name) */
PHP_METHOD(dir, __construct);
PHP_METHOD(dir, rewind);
PHP_METHOD(dir, valid);
PHP_METHOD(dir, key);
PHP_METHOD(dir, current);
PHP_METHOD(dir, next);
PHP_METHOD(dir, getPath);
PHP_METHOD(dir, getFilename);
PHP_METHOD(dir, isDot);
PHP_METHOD(dir, isDir);


/* declare method parameters */
/* supply a name and default to call by parameter */
static
ZEND_BEGIN_ARG_INFO(arginfo_dir___construct, 0) 
	ZEND_ARG_INFO(0, path)  /* parameter name */
ZEND_END_ARG_INFO();


/* the method table */
/* each method can have its own parameters and visibility */
static zend_function_entry util_dir_class_functions[] = {
	PHP_ME(dir, __construct,   arginfo_dir___construct, ZEND_ACC_PUBLIC)
	PHP_ME(dir, rewind,        NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, valid,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, key,           NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, current,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, next,          NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, getPath,       NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, getFilename,   NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, isDot,         NULL, ZEND_ACC_PUBLIC)
	PHP_ME(dir, isDir,         NULL, ZEND_ACC_PUBLIC)
	{NULL, NULL, NULL}
};


/* declare the class handlers */
static zend_object_handlers util_dir_handlers;


/* decalre the class entry */
static zend_class_entry *util_ce_dir;


/* the overloaded class structure */

/* overloading the structure results in the need of having 
   dedicated creatin/cloning/destruction functions */
typedef struct _util_dir_object {
	zend_object       std;
	php_stream        *dirp;
	php_stream_dirent entry;
	char              *path;
	int               index;
} util_dir_object;


/* {{{ util_dir_object_free_storage */
/* close all resources and the memory allocated for the object */
static void util_dir_object_free_storage(void *object TSRMLS_DC)
{
	util_dir_object *intern = (util_dir_object *)object;

	zend_hash_destroy(intern->std.properties);
	FREE_HASHTABLE(intern->std.properties);

	if (intern->path) {
		efree(intern->path);
	}
	if (intern->dirp) {
		php_stream_close(intern->dirp);
	}
	efree(object);
}
/* }}} */


/* {{{ util_dir_object_new */
/* creates the object by 
   - allocating memory 
   - initializing the object members
   - storing the object
   - setting it's handlers

   called from 
   - clone
   - new
 */
static zend_object_value util_dir_object_new_ex(zend_class_entry *class_type, util_dir_object **obj TSRMLS_DC)
{
	zend_object_value retval;
	util_dir_object *intern;
	zval *tmp;

	intern = emalloc(sizeof(util_dir_object));
	memset(intern, 0, sizeof(util_dir_object));
	intern->std.ce = class_type;
	*obj = intern;

	ALLOC_HASHTABLE(intern->std.properties);
	zend_hash_init(intern->std.properties, 0, NULL, ZVAL_PTR_DTOR, 0);
	zend_hash_copy(intern->std.properties, &class_type->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(intern, NULL, (zend_objects_free_object_storage_t) util_dir_object_free_storage, NULL TSRMLS_CC);
	retval.handlers = &util_dir_handlers;
	return retval;
}
/* }}} */


/* {{{ util_dir_object_new */
/* See util_dir_object_new_ex */
static zend_object_value util_dir_object_new(zend_class_entry *class_type TSRMLS_DC)
{
	util_dir_object *tmp;
	return util_dir_object_new_ex(class_type, &tmp TSRMLS_CC);
}
/* }}} */


/* {{{ util_dir_open */
/* open a directory resource */
static void util_dir_open(util_dir_object* intern, char *path TSRMLS_DC)
{
	intern->dirp = php_stream_opendir(path, ENFORCE_SAFE_MODE|REPORT_ERRORS, NULL);

	intern->path = estrdup(path);
	intern->index = 0;

	if (intern->dirp == NULL) {
		/* throw exception: should've been already happened */
		intern->entry.d_name[0] = '\0';
	} else {
		if (!php_stream_readdir(intern->dirp, &intern->entry)) {
			intern->entry.d_name[0] = '\0';
		}
	}
}
/* }}} */

/* {{{ util_dir_object_clone */
/* Local zend_object_value creation (on stack)
   Load the 'other' object 
   Create a new empty object (See util_dir_object_new_ex)
   Open the directory
   Clone other members (properties)
 */
static zend_object_value util_dir_object_clone(zval *zobject TSRMLS_DC)
{
	zend_object_value new_obj_val;
	zend_object *old_object;
	zend_object *new_object;
	zend_object_handle handle = Z_OBJ_HANDLE_P(zobject);
	util_dir_object *intern;

	old_object = zend_objects_get_address(zobject TSRMLS_CC);
	new_obj_val = util_dir_object_new_ex(old_object->ce, &intern TSRMLS_CC);
	new_object = &intern->std;

	util_dir_open(intern, ((util_dir_object*)old_object)->path TSRMLS_CC);

	zend_objects_clone_members(new_object, new_obj_val, old_object, handle TSRMLS_CC);

	return new_obj_val;
}
/* }}} */

/* {{{ util_dir_get_ce */
#if MBO_0
static zend_class_entry *util_dir_get_ce(zval *object TSRMLS_DC)
{
	return util_ce_dir;
}
#endif
/* }}} */

#if PHP_MAJOR_VERSION >= 6 | (PHP_MAJOR_VERSION == 5 & PHP_MINOR_VERSION >= 2)
#define EXTSRM TSRMLS_C
#else
#define EXTSRM
#endif

/* {{{ proto void dir::__construct(string path)
 Cronstructs a new dir iterator from a path. */
/* php_set_error_handling() is used to throw exceptions in case
   the constructor fails. Here we use this to ensure the object
   has a valid directory resource.
   
   When the constructor gets called the object is already created 
   by the engine, so we must only call 'additional' initializations.
 */
PHP_METHOD(dir, __construct)
{
	zval *object = getThis();
	util_dir_object *intern;
	char *path;
	long len;

	php_set_error_handling(EH_THROW, zend_exception_get_default(EXTSRM) TSRMLS_CC);

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &path, &len) == FAILURE) {
		php_set_error_handling(EH_NORMAL, NULL TSRMLS_CC);
		return;
	}

	intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);
	util_dir_open(intern, path TSRMLS_CC);

	php_set_error_handling(EH_NORMAL, NULL TSRMLS_CC);
}
/* }}} */

/* {{{ proto void dir::rewind()
   Rewind dir back to the start */
PHP_METHOD(dir, rewind)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	intern->index = 0;
	if (intern->dirp) {
		php_stream_rewinddir(intern->dirp);
	}
	if (!intern->dirp || !php_stream_readdir(intern->dirp, &intern->entry)) {
		intern->entry.d_name[0] = '\0';
	}
}
/* }}} */

/* {{{ proto string dir::key()
   Return current dir entry */
PHP_METHOD(dir, key)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	if (intern->dirp && intern->entry.d_name[0] != '\0') {
		char *filename;
		long filename_len = spprintf(&filename, 0, "%s/%s", intern->path, intern->entry.d_name);
		RETURN_STRINGL(filename, filename_len, 0);
	} else {
		RETURN_FALSE;
	}
}
/* }}} */

/* {{{ proto string|false dir::current()
   Return this (needed for Iterator interface) */
PHP_METHOD(dir, current)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	if (intern->entry.d_name[0]) {
		RETURN_STRING(intern->entry.d_name, 1);
	} else {
		RETURN_FALSE;
	}
}
/* }}} */

/* {{{ proto void dir::next()
   Move to next entry */
PHP_METHOD(dir, next)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	intern->index++;
	if (!intern->dirp || !php_stream_readdir(intern->dirp, &intern->entry)) {
		intern->entry.d_name[0] = '\0';
	}
}
/* }}} */

/* {{{ proto string dir::valid()
   Check whether dir contains more entries */
PHP_METHOD(dir, valid)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	RETURN_BOOL(intern->entry.d_name[0] != '\0');
}
/* }}} */

/* {{{ proto string dir::getPath()
   Return directory path */
PHP_METHOD(dir, getPath)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	RETURN_STRING(intern->path, 1);
}
/* }}} */

/* {{{ proto string dir::getFilename()
   Return filename of current dir entry */
PHP_METHOD(dir, getFilename)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	RETURN_STRING(intern->entry.d_name, 1);
}
/* }}} */

/* {{{ proto string dir::isDot()
   Returns true if current entry is '.' or  '..' */
PHP_METHOD(dir, isDot)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	RETURN_BOOL(!strcmp(intern->entry.d_name, ".") || !strcmp(intern->entry.d_name, ".."));
}
/* }}} */

/* {{{ proto string dir::isDir()
   Returns whether current entry is a directory  */
PHP_METHOD(dir, isDir)
{
	zval *object = getThis();
	util_dir_object *intern = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);
	char *filename;
	
	php_stat_len filename_len = spprintf(&filename, 0, "%s/%s", intern->path, intern->entry.d_name);

	php_stat(filename, filename_len, /*funcnum*/FS_IS_DIR, return_value TSRMLS_CC);

	efree(filename);
}
/* }}} */

/* define an overloaded iterator structure */
typedef struct {
	zend_object_iterator  intern;
	zval                  *current;
	util_dir_object       *object;
} util_dir_it;

/* {{{ util_dir_it_invalidate_current */
static void util_dir_it_invalidate_current(zend_object_iterator *iter TSRMLS_DC)
{
	util_dir_it *iterator = (util_dir_it *)iter;

	if (iterator->current) {
		zval_ptr_dtor(&iterator->current);
		iterator->current = NULL;
	}
}
/* }}} */
	

/* {{{ util_dir_it_dtor */
static void util_dir_it_dtor(zend_object_iterator *iter TSRMLS_DC)
{
	util_dir_it *iterator = (util_dir_it *)iter;
	zval        *intern = (zval*)iterator->intern.data;

	util_dir_it_invalidate_current(iter TSRMLS_CC);
	zval_ptr_dtor(&intern);

	efree(iterator);
}
/* }}} */
	

/* {{{ util_dir_it_has_more */
static int util_dir_it_has_more(zend_object_iterator *iter TSRMLS_DC)
{
	util_dir_it       *iterator = (util_dir_it *)iter;
	util_dir_object   *object   = iterator->object;

	return object->entry.d_name[0] != '\0' ? SUCCESS : FAILURE;
}
/* }}} */


/* {{{ util_dir_it_current_data */
static void util_dir_it_current_data(zend_object_iterator *iter, zval ***data TSRMLS_DC)
{
	util_dir_it       *iterator = (util_dir_it *)iter;
	util_dir_object   *object   = iterator->object;

	if (!object->dirp || object->entry.d_name[0] == '\0') {
		ALLOC_INIT_ZVAL(iterator->current);
	} else {
		MAKE_STD_ZVAL(iterator->current);
		ZVAL_STRING(iterator->current, object->entry.d_name, 1);
	}
	*data = &iterator->current;
}
/* }}} */


/* {{{ util_dir_it_current_key */
static int util_dir_it_current_key(zend_object_iterator *iter, char **str_key, uint *str_key_len, ulong *int_key TSRMLS_DC)
{
	util_dir_it       *iterator = (util_dir_it *)iter;
	util_dir_object   *object   = iterator->object;

	if (object->dirp && object->entry.d_name[0] != '\0')
	{
		*str_key_len = spprintf(str_key, 0, "%s/%s", object->path, object->entry.d_name);
		(*str_key_len)++;
		return HASH_KEY_IS_STRING;
	}
	else
	{
		*int_key = -1;
		return HASH_KEY_IS_LONG;
	}
}
/* }}} */


/* {{{ util_dir_it_move_forward */
static void util_dir_it_move_forward(zend_object_iterator *iter TSRMLS_DC)
{
	util_dir_it       *iterator = (util_dir_it *)iter;
	util_dir_object   *object   = iterator->object;
	
	object->index++;
	util_dir_it_invalidate_current(iter TSRMLS_CC);
	if (!object->dirp || !php_stream_readdir(object->dirp, &object->entry)) {
		object->entry.d_name[0] = '\0';
	}
}
/* }}} */


/* {{{ util_dir_it_rewind */
static void util_dir_it_rewind(zend_object_iterator *iter TSRMLS_DC)
{
	util_dir_it       *iterator = (util_dir_it *)iter;
	util_dir_object   *object   = iterator->object;
	
	object->index = 0;
	if (object->dirp) {
		php_stream_rewinddir(object->dirp);
	}
	util_dir_it_invalidate_current(iter TSRMLS_CC);
	if (!object->dirp || !php_stream_readdir(object->dirp, &object->entry)) {
		object->entry.d_name[0] = '\0';
	}
}
/* }}} */


/* iterator handler table */
zend_object_iterator_funcs util_dir_it_funcs = {
	util_dir_it_dtor,
	util_dir_it_has_more,
	util_dir_it_current_data,
	util_dir_it_current_key,
	util_dir_it_move_forward,
	util_dir_it_rewind,
	util_dir_it_invalidate_current
};


/* {{{ util_dir_get_iterator */
zend_object_iterator *util_dir_get_iterator(zend_class_entry *ce, zval *object TSRMLS_DC)
{
	util_dir_it       *iterator   = emalloc(sizeof(util_dir_it));
	util_dir_object   *dir_object = (util_dir_object*)zend_object_store_get_object(object TSRMLS_CC);

	object->refcount++;
	iterator->intern.data = (void*)object;
	iterator->intern.funcs = &util_dir_it_funcs;
	MAKE_STD_ZVAL(iterator->current);
	iterator->object = dir_object;
	iterator->current = NULL;
	
	return (zend_object_iterator*)iterator;
}
/* }}} */


/* {{{ proto bool util_invert(bool b)
 Invert a boolean parameter */
PHP_FUNCTION(util_invert)
{
	zend_bool b;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
			"b", &b) == FAILURE) {
		return;
	}

	b = b ? 0 : 1;
	
	RETURN_BOOL(b);
}
/* }}} */

/* {{{ proto bool util_increment(int v [, int max])
 Increment a value with optional maximum */
PHP_FUNCTION(util_increment)
{
	long l, lmax = LONG_MAX;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
			"l|l", &l, &lmax) == FAILURE) {
		RETURN_FALSE;
	}

	l = (l+1) % lmax;
	
	RETURN_LONG(l);
}
/* }}} */

#define UTIL_VERSION_MAJOR	1
#define UTIL_VERSION_MINOR	1

/* {{{ proto bool util_version()
 Retrieve util version */
PHP_FUNCTION(util_version)
{
	char * ver;
	int len;

	len = spprintf(&ver, 0, "%d.%d (%s)", 
		UTIL_VERSION_MAJOR, UTIL_VERSION_MINOR, "$Id: $");
	
	RETURN_STRINGL(ver, len, 0);
}
/* }}} */

/* {{{ proto bool util_version_array()
 Retrieve yourext version as array */
PHP_FUNCTION(util_version_array)
{
	char *ver;
	int len = spprintf(&ver, 0, "%d.%d", 
		UTIL_VERSION_MAJOR, UTIL_VERSION_MINOR);
	
	array_init(return_value);
	add_assoc_long(return_value, "major", UTIL_VERSION_MAJOR);
	add_assoc_long(return_value, "minor", UTIL_VERSION_MINOR);
	add_assoc_string(return_value, "cvs", "$Id: $", 1);
	add_assoc_stringl(return_value, "ver", ver, len, 0);
}
/* }}} */

/* {{{ */
int util_foreach(zval **param, zval *func_name TSRMLS_DC)
{
	zval retval;
	zval *args[1];
	int status;

	args[0] = *param;
	status = call_user_function(EG(function_table), NULL, 
		func_name, &retval, 1, args TSRMLS_CC);

	if (!zend_is_true(&retval)) status = FAILURE;
	zval_dtor(&retval);		

	return status == SUCCESS
			? ZEND_HASH_APPLY_KEEP 
			: ZEND_HASH_APPLY_STOP;
} /* }}} */

/* {{{ proto void util_foreach(array ar, mixed func)
  Call a function for all elements: bool apply(mixed param) */
PHP_FUNCTION(util_foreach)
{
	zval *ar, *zfunc;
	char *fname;
	
	if ((zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET,
 	  	ZEND_NUM_ARGS() TSRMLS_CC, "az", &ar, &zfunc) 
		== FAILURE &&
	  zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "oz", &ar,&zfunc)
		 == FAILURE) 
	  || !zend_is_callable(zfunc, 0, &fname)) {
		return;
	}
	zend_hash_apply_with_argument(HASH_OF(ar), 
		(apply_func_arg_t)util_foreach, zfunc TSRMLS_CC);
} /* }}} */

/* If you declare any globals in php_util.h uncomment this:
ZEND_DECLARE_MODULE_GLOBALS(util)
*/

/* {{{ util_functions[]
 *
 * Every user visible function must have an entry in util_functions[].
 */
function_entry util_functions[] = {
	PHP_FE(util_invert,        NULL)
	PHP_FE(util_increment,     NULL)
	PHP_FE(util_version,       NULL)
	PHP_FE(util_version_array, NULL)
	PHP_FE(util_foreach,       NULL)
	{NULL, NULL, NULL}	/* Must be the last line in util_functions[] */
};
/* }}} */

/* {{{ util_module_entry
 */
zend_module_entry util_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"util",
	util_functions,
	PHP_MINIT(util),
	PHP_MSHUTDOWN(util),
	PHP_RINIT(util),		/* Replace with NULL if there's nothing to do at request start */
	PHP_RSHUTDOWN(util),	/* Replace with NULL if there's nothing to do at request end */
	PHP_MINFO(util),
#if ZEND_MODULE_API_NO >= 20010901
	"0.1", /* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_UTIL
ZEND_GET_MODULE(util)
#endif

/* {{{ PHP_INI
 */
/* Remove comments and fill if you need to have entries in php.ini
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("util.global_value",      "42", PHP_INI_ALL, OnUpdateLong, global_value, zend_util_globals, util_globals)
    STD_PHP_INI_ENTRY("util.global_string", "foobar", PHP_INI_ALL, OnUpdateString, global_string, zend_util_globals, util_globals)
PHP_INI_END()
*/
/* }}} */

/* {{{ php_util_init_globals
 */
/* Uncomment this function if you have INI entries
static void php_util_init_globals(zend_util_globals *util_globals)
{
	util_globals->global_value = 0;
	util_globals->global_string = NULL;
}
*/
/* }}} */

/* {{{ PHP_MINIT_FUNCTION(util) */
PHP_MINIT_FUNCTION(util)
{
	zend_class_entry ce;

	INIT_CLASS_ENTRY(ce, "dirs", util_dir_class_functions);
	util_ce_dir = zend_register_internal_class(&ce TSRMLS_CC);
	util_ce_dir->create_object = util_dir_object_new;
	zend_class_implements(util_ce_dir TSRMLS_CC, 1, zend_ce_iterator);
	memcpy(&util_dir_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
	util_dir_handlers.clone_obj = util_dir_object_clone;

	util_ce_dir->get_iterator = util_dir_get_iterator;

	/* If you have INI entries, uncomment these lines 
	ZEND_INIT_MODULE_GLOBALS(util, php_util_init_globals, NULL);
	REGISTER_INI_ENTRIES();
	*/
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(util)
{
	/* uncomment this line if you have INI entries
	UNREGISTER_INI_ENTRIES();
	*/
	return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request start */
/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(util)
{
	return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request end */
/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(util)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(util)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "util support", "enabled");
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}
/* }}} */


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
